"""Caching interface to Kerr QNMs

This is a high-level interface to the package.
The global cache :obj:`qnm.modes_cache` (an instance of
:class:`KerrSeqCache`) will return instances of
:class:`qnm.spinsequence.KerrSpinSeq` from memory or disk. If a spin
sequence is neither in memory nor on disk then it will be computed and
returned.

Use :meth:`download_data` to fetch a collection of precomputed spin
sequences from the web.
"""

from __future__ import division, print_function, absolute_import

import logging
import pickle
import os
import sys
try:
    from pathlib import Path # py 3
except ImportError:
    from pathlib2 import Path # py 2
try:
    from urllib.request import urlretrieve # py 3
except ImportError:
    from urllib import  urlretrieve # py 2
from tqdm import tqdm
import tarfile
import glob

import numpy as np

from .angular import l_min
from .spinsequence import KerrSpinSeq
from .schwarzschild.tabulated import QNMDict

# TODO should all the functions be static member functions? No, I don't
# think so

# Taken from matplotlib
def get_home():
    """
    Return the user's home directory.
    If the user's home directory cannot be found, return None.
    """
    try:
        return str(Path.home())
    except Exception:
        return None

def get_cachedir():
    """
    Return the location of the cache directory.  This follows a
    pattern similar to matplotlib's treatment of config/cache dirs.

    The directory is chosen as follows:
    1. If the QNMCACHEDIR environment variable is supplied, choose that.
    2a. On Linux, follow the XDG specification and look first in
        `$XDG_CACHE_HOME`, if defined, or `$HOME/.cache`.
    2b. On other platforms, choose `$HOME/.qnm`.
    3. If the chosen directory exists and is writable, use that as the
       configuration directory.
    4. A writable directory could not be found; return None.

    Returns
    -------
    pathlib.Path object or None
    """
    xdg_cache_dir = (os.environ.get('XDG_CACHE_HOME')
                     or (str(Path(get_home(), ".cache"))
                         if get_home()
                         else None))

    cachedir = os.environ.get('QNMCACHEDIR')
    if cachedir:
        cachedir = Path(cachedir).resolve()
    elif sys.platform.startswith(('linux', 'freebsd')) and xdg_cache_dir:
        cachedir = Path(xdg_cache_dir, "qnm")
    elif get_home():
        cachedir = Path(get_home(), ".qnm")
    else:
        cachedir = None

    if cachedir:
        try:
            cachedir.mkdir(parents=True, exist_ok=True)
        except OSError:
            pass
        else:
            if os.access(str(cachedir), os.W_OK) and cachedir.is_dir():
                return cachedir

    # If all else fails...
    return None

def mode_pickle_path(s, l, m, n):
    """Construct the path to a pickle file for the mode (s, l, m, n)

    Parameters
    ----------
    s: int
      Spin-weight of field of interest.

    l: int
      Multipole number of mode. l >= angular.l_min(s, m)

    m: int
      Azimuthal number of mode.

    n: int
      Overtone number of mode.

    Returns
    -------
    pathlib.Path object or None
      `<cachedir>/data/s<s>_l<l>_m<m>_n<n>.pickle`

     """

    if not (l >= l_min(s, m)):
        raise ValueError("l={} must be >= l_min={}"
                         .format(l, l_min(s, m)))


    s_sign = '-' if (s<0) else ''
    m_sign = '-' if (m<0) else ''

    filename = 's{}{}_l{}_m{}{}_n{}.pickle'.format(
        s_sign, np.abs(s), l,
        m_sign, np.abs(m), n)

    cache_dir = get_cachedir()
    if cache_dir is not None:
        pickle_path = get_cachedir() / 'data' / filename
        return pickle_path
    else:
        return None

def write_mode(spin_seq, pickle_path=None):
    """Write an instance of KerrSpinSeq to disk.

    Parameters
    ----------
    spin_seq: KerrSpinSeq
      The mode to write to disk.

    pickle_path: string or pathlib.Path, optional [default: None]
      Path to file to write. If None, get the path from
      :meth:`mode_pickle_path()`.

    Raises
    ------
    TODO

    """

    if (pickle_path is None):
        pickle_path = mode_pickle_path(spin_seq.s, spin_seq.l,
                                       spin_seq.m, spin_seq.n)
        if pickle_path is None:
            logging.error('No cache dir found, not writing anything.')
            return

    # Convert pickle_path to Path if it's a string
    pickle_path = Path(pickle_path)
    the_dir = pickle_path.parent
    if not the_dir.exists():
        try:
            the_dir.mkdir(parents=True, exist_ok=True)
        except:
            logging.error("Could not create dir {} to store Kerr QNM sequence".format(the_dir))

    try:
        with pickle_path.open('wb') as handle:
            logging.info("Writing Kerr QNM sequence to file {}".format(pickle_path))
            pickle.dump(spin_seq, handle, protocol=pickle.HIGHEST_PROTOCOL)
    except:
        logging.error("Could not write Kerr QNM sequence to file {}".format(pickle_path))

    return

def load_cached_mode(s, l, m, n):
    """Read a KerrSpinSeq from disk.

    Path is determined by :meth:`mode_pickle_path(s, l, m, n)`.

    Parameters
    ----------
    s: int
      Spin-weight of field of interest.

    l: int
      Multipole number of mode. l >= angular.l_min(s, m)

    m: int
      Azimuthal number of mode.

    n: int
      Overtone number of mode.

    Returns
    -------
    KerrSpinSeq
      The mode, if it exists. Otherwise None.

    """

    spin_seq = None

    pickle_path = mode_pickle_path(s, l, m, n)

    if (pickle_path is None) or (not pickle_path.exists()):
        return None

    try:
        with pickle_path.open('rb') as handle:
            logging.info("Loading Kerr QNM sequence from file {}".format(pickle_path))
            spin_seq = pickle.load(handle)
    except UnicodeDecodeError as e:
        with pickle_path.open('rb') as handle:
            logging.info("Loading Kerr QNM sequence from file {}".format(pickle_path))
            spin_seq = pickle.load(handle, encoding='latin1')
    except:
        logging.warn("Could not load Kerr QNM sequence from file {}.".format(pickle_path))
        if not hasattr(load_cached_mode, "have_warned"):
            logging.warn("Do you need to run qnm.download_data()?")
            load_cached_mode.have_warned = True

    return spin_seq

############################################################

class KerrSeqCache(object):
    """High-level caching interface for getting precomputed spin sequences.

    An instance of :class:`KerrSeqCache` will return instances of
    :class:`qnm.spinsequence.KerrSpinSeq` from memory or disk. If a
    spin sequence is neither in memory nor on disk then it will be
    computed and returned.

    Use :meth:`download_data` to fetch a collection of precomputed
    spin sequences from the web.

    Parameters
    ----------
    init_schw: bool, optional [default: False]
      Value of init flag to pass to
      :class:`qnm.schwarzschild.tabulated.QNMDict`. You should set this to
      True the first time in a session that you create a QNMDict
      (most likely via this class).

    compute_if_not_found: bool, optional [default: True]
      If a mode sequence is not found on disk, this flag controls
      whether to try to compute the sequence from scratch.

    compute_pars: dict, optional [default: None]
      A dict of parameters to pass to
      :class:`qnm.spinsequence.KerrSpinSeq` if computing a mode
      sequence from scratch.

    Examples
    --------

    >>> import qnm
    >>> # qnm.download_data() # Only need to do this once
    >>> grav_220 = qnm.modes_cache(s=-2,l=2,m=2,n=0)
    >>> omega, A, C = grav_220(a=0.68)
    >>> print(omega)
    (0.5239751042900845-0.08151262363119986j)

    """

    # Borg pattern, the QNM dict will be shared among all instances
    _shared_state = {}

    def __init__(self, init_schw=False, compute_if_not_found=True,
                 compute_pars=None):

        self.__dict__ = self._shared_state

        # Reset these attributes
        # TODO Is this the right thing to do? Maybe not?
        self.compute_if_not_found = compute_if_not_found
        self.compute_pars = dict(compute_pars) if compute_pars is not None else {}

        # Don't reset this one
        if (not hasattr(self, 'schw_dict')):
            # First!
            # The only reason for keeping this around is to know
            # whether or not QNMDict has been initialized
            # TODO FIGURE OUT IF WE SHOULD DO THIS
            self.schw_dict = QNMDict(init=init_schw)

        # Definitely don't reset this one
        if (not hasattr(self, 'seq_dict')):
            # First!
            self.seq_dict = {}


    def __call__(self, s, l, m, n,
                 compute_if_not_found=None,
                 compute_pars=None):
        """Load a :class:`qnm.spinsequence.KerrSpinSeq` from the cache
        or from disk if available.

        If the mode sequence is not available on disk, optionally
        compute it from scratch.

        Parameters
        ----------
        s: int
          Spin-weight of field of interest.

        l: int
          Multipole number of mode. l >= angular.l_min(s, m)

        m: int
          Azimuthal number of mode.

        n: int
          Overtone number of mode.

        compute_if_not_found: bool, optional [default: self.compute_if_not_found]
          Whether or not to compute from scratch the spin sequence if
          it is not available on disk.

        compute_pars: dict, optional [default: self.compute_pars]
          Dict of parameters to pass to KerrSpinSeq if a mode sequence
          needs to be computed from scratch.

        Returns
        -------
        KerrSpinSeq
          The mode, if it is in the cache, on disk, or has been
          computed from scratch.  If the mode is not available and
          compute_if_not_found is false, return None.

        """

        if not (l >= l_min(s, m)):
            raise ValueError("l={} must be >= l_min={}"
                             .format(l, l_min(s, m)))

        if not (n >= 0):
            raise ValueError("n={} must be non-negative".format(n))

        if compute_if_not_found is None:
            compute_if_not_found = self.compute_if_not_found

        if compute_pars is None:
            compute_pars = self.compute_pars

        key = (s,l,m,n)
        if (key in self.seq_dict.keys()):
            return self.seq_dict[key]

        loaded_mode = load_cached_mode(s,l,m,n)

        if (loaded_mode is not None):
            self.seq_dict[key] = loaded_mode
            return loaded_mode

        if (compute_if_not_found):
            logging.info("Mode not in cache, or on disk, trying to "
                         "compute s={}, l={}, m={}, n={}"
                         .format(s, l, m, n))
            the_pars = dict(compute_pars) # Make a copy
            the_pars.update({ 's': s, 'l': l, 'm': m, 'n': n })
            computed = KerrSpinSeq(**the_pars)
            computed.do_find_sequence()
            self.seq_dict[key] = computed
            return computed

        return None

    def write_all(self):
        """Write all of the modes in the cache to disk.

        TODO: Take an overwrite argument which will force overwrite or
        not.
        """

        for _, seq in self.seq_dict.items():
            write_mode(seq)

############################################################
def build_package_default_cache(ksc):
    """Compute the standard list of modes that this package
    promises to have in its cache.

    This method is intended to be used for building the modes from
    scratch in a predictable way.  If modes are available on disk then
    there will be no computation, simply loading all the default modes.

    Parameters
    ----------
    ksc: KerrSeqCache
      The cache that will hold the modes we are about to compute.

    Returns
    -------
    KerrSeqCache
      The updated cache.
    """

    QNMDict(init=True)

    a_max   = .9995
    tol     = 1e-10
    cf_tol  = np.sqrt(np.finfo(float).eps)
    delta_a = 2.5e-3
    Nr_max  = 6000

    ksc.compute_pars.update({'a_max': a_max, 'tol': tol, 'cf_tol': cf_tol,
                             'delta_a': delta_a, 'Nr_max': Nr_max})

    # Known modes that become algebraically special
    modes_to_avoid = [(-1, 1, 0, 6)]

    reruns = []

    ns=np.arange(0,7)
    ss = [-2, -1]
    for s in ss:
        ls=np.arange(np.abs(s),8)
        for l in ls:
            ms=np.arange(-l,l+1)
            for m in ms:
                for n in ns:
                    if ((s, l, m, n) in modes_to_avoid):
                        print("Skipping known mode {}".format((s, l, m, n)))
                        continue
                    try:
                        ksc(s, l, m, n)
                    except:
                        reruns.append((s,l,m,n))

    print('{} modes in cache, going to rerun {} modes'.format(len(ksc.seq_dict),
                                                              len(reruns)))

    # Tweak the params a little bit for the reruns
    re2runs = []
    ksc.compute_pars.update({'delta_a': 1.9e-3, 'cf_tol': cf_tol * 0.25, 'tol': tol * 3.})
    for s, l, m, n in reruns:
        try:
            ksc(s, l, m, n)
        except:
            re2runs.append((s, l, m, n))

    print('{} modes in cache, going to rerun {} modes'.format(len(ksc.seq_dict),
                                                              len(re2runs)))

    ksc.compute_pars.update({'a_max': .9992, 'delta_a': 1.7e-3,
                             'cf_tol': cf_tol * 0.05, 'tol': tol * 10.,
                             'Nr_max': Nr_max * 2})

    re3runs = []
    for s, l, m, n in re2runs:
        try:
            ksc(s, l, m, n)
        except:
            re3runs.append((s, l, m, n))

    print('{} modes in cache, going to rerun {} modes'.format(len(ksc.seq_dict),
                                                              len(re3runs)))

    ksc.compute_pars.update({'a_max': .9993, 'delta_a': 0.24e-3,
                             'cf_tol': cf_tol * 0.01, 'tol': tol * 70.,
                             'Nr_max': Nr_max * 3})

    # This is the last round of reruns we need to do
    for s, l, m, n in re3runs:
        ksc(s, l, m, n)

    return ksc

############################################################
# This is taken verbatim from the tqdm examples, see
# https://pypi.org/project/tqdm/#hooks-and-callbacks

class _TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""
    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


############################################################

# NOTE This URL is hardcoded. The data version to use may need to
# be updated if fields inside classes change.
_data_url = 'https://duetosymmetry.com/files/qnm/data-0.4.0.tar.bz2'

def download_data(overwrite=False):
    """Fetch and decompress tarball of precomputed spin sequences from
    the web.

    Parameters
    ----------
    overwrite: bool, optional [default: False]
      If there is already a tarball on disk, this flag controls
      whether or not it is overwritten.

    """

    filename = _data_url.split('/')[-1]
    base_dir = get_cachedir()
    if base_dir is not None:
        dest     = base_dir / filename
    else:
        print('No cache dir found, not downloading anything.')
        return

    if (dest.exists() and (overwrite is False)):
        print("Destination path {} already exists, use overwrite=True "
              "to force an overwrite.".format(dest))
        return

    print("Trying to fetch {}".format(_data_url))
    with _TqdmUpTo(unit='B', unit_scale=True, miniters=1,
                   desc=filename) as t:
        urlretrieve(_data_url, filename=str(dest), reporthook=t.update_to)

    _decompress_data()

############################################################
def _decompress_data():
    """Decompress tarball of precomputed spin sequences."""

    dest_dir = get_cachedir()
    if dest_dir is None:
        print('No cache dir found, not decompressing anything.')
        return

    filename = _data_url.split('/')[-1]
    tarball = dest_dir / filename

    print("Trying to decompress file {}".format(tarball))
    with tarfile.open(str(tarball), "r:bz2") as tar:
        tar.extractall(str(dest_dir))

    data_dir = dest_dir / 'data'
    pickle_files = data_dir.glob('*.pickle')
    print("Data directory {} contains {} pickle files"
          .format(data_dir, len(list(pickle_files))))

############################################################
def _clear_disk_cache(delete_tarball=False):
    """Delete disk cache of precomputed spin sequences."""

    base_dir = get_cachedir()

    if base_dir is None:
        print('No cache dir found, not deleting anything.')
        return

    data_dir = base_dir / 'data'
    pickle_files = data_dir.glob('*.pickle')

    for pickle_file in pickle_files:
        try:
            pickle_file.unlink()
        except OSError:
            print('Could not remove file "{}"'.format(pickle_file))

    if delete_tarball:
        tarball_path = base_dir / 'data.tar.bz2'
        try:
            tarball_path.unlink()
        except:
            print('Could not remove file "{}"'.format(tarball_path))



""" Follow a Schwarzschild QNM sequence (s,l) from n=0 upwards.

The class :class:`SchwOvertoneSeq` makes it possible to find
successive overtones (n's) of a QNM labeled by (s,l), in Schwarzschild
(a=0). See its documentation for more details.
"""

from __future__ import division, print_function, absolute_import

import logging

import numpy as np
from scipy import optimize, interpolate

from ..angular import l_min, swsphericalh_A
from ..nearby import NearbyRootFinder
from .approx import dolan_ottewill_expansion

# TODO some documentation here, better documentation throughout

class SchwOvertoneSeq(object):
    """Object to follow a sequence of Schwarzschild overtones,
    starting from n=0.  First two overtone seeds come from
    approx.dolan_ottewill_expansion, and afterwards linear
    extrapolation on the solutions is used to seed the root
    finding for higher values of n. Uses
    :class:`qnm.nearby.NearbyRootFinder` to actually perform the
    root-finding.

    Attributes
    ----------
    A: float
      Value of the angular separation constant.

    n: array of int
      Overtone numbers.

    omega: np.array of complex
      The QNM frequencies along the overtone sequence, element i is
      overtone i.

    cf_err: np.array of float
      Estimate of continued fraction truncation error in solving for
      QNM frequency.

    n_frac: np.array of int
      Truncation number of continued fraction.

    solver: NearbyRootFinder
      Instance of :class:`qnm.nearby.NearbyRootFinder` that is used to
      find the QNMs.

    Parameters
    ----------
    n_max: int [default: 12]
      Maximum overtone number to search for (must be positive).

    s: int [default: 2]
      Spin weight of field of interest.

    [The m parameter is omitted because this is just for Schwarzschild.]

    l: int [default: 2]
      The multipole number of a sequence starting from the
      analytically-known value at a=0 .

    tol: float [default: 1e-10]
      Tolerance for root-finding.

    Nr: int [default: 300]
      Truncation number of radial infinite continued
      fraction. Must be sufficiently large for convergence.

    r_N: complex [default: 0.j]
      Seed value taken for truncation of infinite continued
      fraction.

    Examples
    --------
    Suppose you want the n=5 overtone for (s=-1, l=3):

    >>> from qnm.schwarzschild.overtonesequence import SchwOvertoneSeq
    >>> seq = SchwOvertoneSeq(s=-1, l=3, n_max=5)
    >>> seq.find_sequence()
    >>> "{:.10f}".format(seq.omega[5])
    '0.5039017454-1.1703558890j'

    Later, you want to go out to n=8:

    >>> seq.extend(n_max=8)
    >>> "{:.10f}".format(seq.omega[8])
    '0.4227909294-1.9136575598j'

    """

    def __init__(self, *args, **kwargs):

        # Read args
        self.n_max       = kwargs.get('n_max',       12)
        self.s           = kwargs.get('s',           -2)
        self.l           = kwargs.get('l',           2)
        self.tol         = kwargs.get('tol',         1e-10)
        self.Nr          = kwargs.get('Nr',          300)
        self.Nr_min      = self.Nr
        self.Nr_max      = kwargs.get('Nr_max',      6000)
        self.r_N         = kwargs.get('r_N',         0.j)

        # TODO check that values make sense!!!
        if not (self.l >= l_min(self.s, 0)):
            raise ValueError("l={} must be >= l_min={}"
                             .format(self.l, l_min(self.s, 0)))

        # We know the Schwarzschild separation constant analytically
        self.A = swsphericalh_A(self.s, self.l, 0)

        # Create array of n's and omega's
        self.n      = []
        self.omega  = np.array([], dtype=complex)
        self.cf_err = np.array([])
        self.n_frac  = np.array([])

        # We need and instance of root finder
        self.solver = NearbyRootFinder(s=self.s, m=0,
                                       l_max=self.l + 1,
                                       a=0.,
                                       A_closest_to=self.A,
                                       tol=self.tol,
                                       n_inv=0, Nr=self.Nr,
                                       Nr_max=self.Nr_max,
                                       r_N=self.r_N)

    def find_sequence(self):
        """ Alias for :meth:`extend()` """

        self.extend()

    def extend(self, n_max=None):
        """ Extend the current overtone sequence to a greater n_max.

        Parameters
        ----------
        n_max: int, optional [default: None]
          If is None, use the value of self.n_max .
          If given, set self.n_max to this new value and proceed.

        Raises
        ------
        optimize.nonlin.NoConvergence
          If a root can't be found within a few [TODO make param?]
          attempted inversion numbers.

        """

        if (n_max is not None):
            self.n_max = n_max

        while (len(self.omega) <= self.n_max):

            n = len(self.omega)

            self.solver.clear_results()
            self.solver.set_params(n_inv=n)

            if (n < 2):
                omega_guess = dolan_ottewill_expansion(self.s, self.l, n)
            else:
                # Linearly extrapolate from the last two
                om_m_1 = self.omega[-1]
                om_m_2 = self.omega[-2]

                om_diff = om_m_1 - om_m_2

                # Linearly interpolate
                omega_guess = om_m_1 + om_diff

                if (n > 5):
                    # Check if this difference is greater than the typical spacing
                    # of the last few points
                    typ_sp = np.mean(np.abs(np.diff(self.omega[-6:-1])))

                    if (np.abs(om_diff) > 2. * typ_sp):
                        # It's likely we skipped an overtone.
                        # Average and go back one
                        omega_guess = (om_m_1 + om_m_2)/2.
                        self.solver.set_params(n_inv=n-1)
                        logging.info("Potentially skipped an overtone "
                                     "in the series, trying to go back")

            self.solver.set_params(omega_guess=omega_guess)
            # Try to reject previously-found poles
            self.solver.set_poles(self.omega)

            # Flag: is the continued fraction expansion converging?
            cf_conv = False

            while not cf_conv:

                result = self.solver.do_solve()

                if (result is None):
                    # Potentially try the next inversion number
                    cur_inv_n = self.solver.n_inv
                    cur_inv_n = cur_inv_n + 1
                    self.solver.set_params(n_inv=cur_inv_n)

                    if (np.abs(cur_inv_n - n) > 2):
                        # Got too far away, give up
                        raise optimize.nonlin.NoConvergence('Failed to find '
                                                            'QNM in sequence '
                                                            'at n={}'.format(n))
                    else:
                        logging.info("Trying inversion number {}".format(cur_inv_n))
                        # Try again
                        continue

                # Ensure we start on the "positive frequency"
                # sequence.  This only works for Schwarzscdhild because
                # there the separation constant is real.
                if (np.real(result) < 0):
                    result = -np.conjugate(result)

                cf_err, n_frac = self.solver.get_cf_err()

                # TODO ACTUALLY DO SOMETHING WITH THESE NUMBERS
                cf_conv = True

                if cf_conv:
                    # Done with this value of n

                    logging.info("s={}, l={}, found what I think is "
                                 "n={}, omega={}".format(self.s,
                                                         self.l, n, result))

                    self.n.append(n)

                    self.omega  = np.append(self.omega,  result)
                    self.cf_err = np.append(self.cf_err, cf_err)
                    self.n_frac = np.append(self.n_frac, n_frac)

                    # Make sure we sort properly!
                    ind_sort = np.argsort(-np.imag(self.omega))
                    self.omega  = self.omega[ind_sort]
                    self.cf_err = self.cf_err[ind_sort]
                    self.n_frac = self.n_frac[ind_sort]

                else:
                    # For the next attempt, try starting where we
                    # ended up
                    self.solver.set_params(omega_guess=result)
                    # Now try again, because cf_conv is still False



""" Follow a QNM labeled by (s,l,m,n) as spin varies from a=0 upwards.
"""


from __future__ import division, print_function, absolute_import

import logging

import numpy as np
from scipy import optimize, interpolate

from .angular import l_min, swsphericalh_A, C_and_sep_const_closest
from .nearby import NearbyRootFinder

from .schwarzschild.approx import Schw_QNM_estimate
from .schwarzschild.tabulated import QNMDict

# TODO some documentation here, better documentation throughout

class KerrSpinSeq(object):
    """Object to follow a QNM up a sequence in a, starting from
    a=0. Values for omega and the separation constant from one
    value of a are used to seed the root finding for the next
    value of a, to maintain continuity in a when separation
    constant order can change. Uses NearbyRootFinder to actually
    perform the root-finding.

    Parameters
    ----------
    a_max: float [default: .99]
      Maximum dimensionless spin of black hole for the sequence,
      0 <= a_max < 1.

    delta_a: float [default: 0.005]
      Step size in a for following the sequence from a=0 to a_max

    delta_a_min: float [default: 1.e-5]
      Minimum step size in a.

    delta_a_max: float [default: 4.e-3]
      Maximum step size in a.

    s: int [default: 2]
      Spin of field of interest

    m: int [default: 2]
      Azimuthal number of mode of interest

    l: int [default: 2]
      The l-number of a sequence starting from the
      analytically-known value at a=0

    l_max: int [default: 20]
      Maximum value of l to include in the spherical-spheroidal
      matrix for finding separation constant and mixing
      coefficients. Must be sufficiently larger than l of interest
      that angular spectral method can converge. The number of
      l's needed for convergence depends on a.

    omega_guess: complex [default: from schwarzschild.QNMDict]
      Initial guess of omega for root-finding

    tol: float [default: sqrt(double epsilon)]
      Tolerance for root-finding omega

    cf_tol: float [defailt: 1e-10]
      Tolerance for continued fraction calculation

    n: int [default: 0]
      Overtone number of interest (sets the inversion number for
      infinite continued fraction in Leaver's method)

    Nr: int [default: 300]
      Truncation number of radial infinite continued
      fraction. Must be sufficiently large for convergence.

    Nr_min: int [default: Nr]
      Minimum number of terms for evaluating continued fraction.

    Nr_max: int [default: 4000]
      Maximum number of terms for evaluating continued fraction.

    r_N: complex [default: 0.j]
      Seed value taken for truncation of infinite continued
      fraction. UNUSED, REMOVE

    """

    def __init__(self, *args, **kwargs):

        # Read args
        self.a_max       = kwargs.get('a_max',       0.99)
        # TODO Maybe change this to delta_a0
        self.delta_a     = kwargs.get('delta_a',     0.005)
        self.delta_a_min = kwargs.get('delta_a_min', 1.e-5)
        self.delta_a_max = kwargs.get('delta_a_max', 4.e-3)
        self.s           = kwargs.get('s',           -2)
        self.m           = kwargs.get('m',           2)
        self.l           = kwargs.get('l',           2)
        self.l_max       = kwargs.get('l_max',       20)
        self.tol         = kwargs.get('tol',         np.sqrt(np.finfo(float).eps))
        self.cf_tol      = kwargs.get('cf_tol',      1e-10)
        self.n           = kwargs.get('n',           0)

        if ('omega_guess' in kwargs.keys()):
            self.omega_guess = kwargs.get('omega_guess')
        else:
            qnm_dict = QNMDict()
            self.omega_guess = qnm_dict(self.s, self.l, self.n)[0]

        self.Nr          = kwargs.get('Nr',          300)
        self.Nr_min      = self.Nr
        self.Nr_max      = kwargs.get('Nr_max',      4000)
        self.r_N         = kwargs.get('r_N',         0.j)

        # TODO check that values make sense!!!
        if not (self.a_max < 1.):
            raise ValueError("a_max={} must be < 1.".format(self.a_max))
        if not (self.l >= l_min(self.s, self.m)):
            raise ValueError("l={} must be >= l_min={}".format(
                self.l, l_min(self.s, self.m)))

        # Create array of a's, omega's, and A's
        self.a     = []
        self.omega = []
        self.cf_err= []
        self.n_frac= []
        self.A     = []
        self.C     = []

        self.delta_a_prop = []

        self._interp_o_r = None
        self._interp_o_i = None
        self._interp_A_r = None
        self._interp_A_i = None

        # We need and instance of root finder
        self.solver = NearbyRootFinder(s=self.s, m=self.m,
                                       l_max=self.l_max,
                                       tol=self.tol,
                                       n_inv=self.n, Nr=self.Nr,
                                       Nr_min=self.Nr_min,
                                       Nr_max=self.Nr_max,
                                       r_N=self.r_N)

    # Change this to *extending* the sequence so this can be reused
    def do_find_sequence(self):
        """Solve for the "spin sequence", i.e. solve for the QNM as we
        go up in spin."""

        logging.info("l={}, m={}, n={} starting".format(
            self.l, self.m, self.n))
        
        i  = 0  # TODO Allow to start at other values
        _a = 0. # TODO Allow to start at other values

        # This flag is set to true if we are trying to rerun because
        # we found Re[\omega] < 0. We will only try this once. If it
        # still comes back as negative, continue?
        om_re_neg_rerun = False

        # Flag about whether we've warned re: imag axis
        warned_imag_axis = False

        # Initializing the sequence, start with guesses
        A0 = swsphericalh_A(self.s, self.l, self.m)
        omega_guess = self.omega_guess

        while _a <= self.a_max:

            self.solver.set_params(a=_a, A_closest_to=A0,
                                   omega_guess=omega_guess)

            result = self.solver.do_solve()

            if (result is None):
                raise optimize.nonlin.NoConvergence('Failed to find '
                                                    'QNM in sequence '
                                                    'at a={}'.format(_a))

            # TODO This probably doesn't belong here
            # Ensure we start on the "positive frequency"
            # sequence.  This only works for i==0 (a=0.) because
            # there the separation constant is real.
            if ((i == 0) and (np.real(result) < 0)):
                result = -np.conjugate(result)
            # if (np.real(result) < 0):
            #     if (not om_re_neg_rerun):
            #         om_re_neg_rerun = True
            #         omega_guess = -np.conj(result)
            #         A0          = np.conj(self.solver.A)
            #         print("neg, retry with om={}, A={}".format(omega_guess,A0))
            #         continue
            #     else:
            #         print("neg, don't retry")
            #         om_re_neg_rerun = False
            #         # give up, don't try again

            # TODO Behavior based on these numbers? Should we continue
            # even if cf_err is larger than the desired error tol?
            cf_err, n_frac = self.solver.get_cf_err()

            # Warn if we're very close to the imaginary axis
            if ((np.abs(np.real(result)) < self.tol)
                and not warned_imag_axis):
                logging.warn("Danger! At a={}, found Re[omega]={}, "
                             "twithin tol={} of the imaginary axis. "
                             "this mode may become algebraically "
                             "special, where Leaver's method would "
                             "fail, or the mode may even disappear."
                             .format(_a, result, self.tol))
                warned_imag_axis = True

            # Done with this value of a
            self.a.append(_a)
            self.omega.append(result)
            self.A.append(self.solver.A)
            self.C.append(self.solver.C)
            self.cf_err.append(cf_err)
            self.n_frac.append(n_frac)

            # We always try to get the a_max value. If that's the
            # value we just did, break out of the loop by hand
            if (_a == self.a_max):
                break

            _a, omega_guess, A0 = self._propose_next_a_om_A()

            i = i+1
            # Go to the next iteration of the _a loop

        logging.info("s={}, l={}, m={}, n={} completed with {} points".format(
            self.s, self.l, self.m, self.n, len(self.a)))

        # Done extending the spin sequence!

        self.build_interps()

    def _propose_next_a_om_A(self):
        """This is an internal function that's used by
        do_find_sequence to compute starting values for the next
        step along the spin sequence."""

        # For the next value of a, start with a guess based on
        # the previously-computed values. When we have two or more
        # values, we can do a quadratic fit. Otherwise just start
        # at the same value.
        _a = self.a[-1]

        if (len(self.a) < 3):
            omega_guess = self.omega[-1]
            A0          = self.A[-1]

            _a = _a + self.delta_a
        else:

            # Build interpolants and allow extrapolation
            # Sadly, UnivariateSpline does not work on complex data
            interp_o_r = interpolate.UnivariateSpline(
                self.a[-3:], np.real(self.omega[-3:]),
                s=0, # No smoothing!
                k=2, ext=0)

            interp_o_i = interpolate.UnivariateSpline(
                self.a[-3:], np.imag(self.omega[-3:]),
                s=0, # No smoothing!
                k=2, ext=0)

            interp_A_r = interpolate.UnivariateSpline(
                self.a[-3:], np.real(self.A[-3:]),
                s=0, # No smoothing!
                k=2, ext=0)

            interp_A_i = interpolate.UnivariateSpline(
                self.a[-3:], np.imag(self.A[-3:]),
                s=0, # No smoothing!
                k=2, ext=0)

            if (True or (len(self.a) > np.Inf)): # Only do the curvature estimate after a while
                # Their second derivatives
                d2_o_r = interp_o_r.derivative(2)
                d2_o_i = interp_o_i.derivative(2)
                d2_A_r = interp_A_r.derivative(2)
                d2_A_i = interp_A_i.derivative(2)

                # Estimate the a-curvature of the omega and A functions:
                d2_o = np.abs(d2_o_r(_a) + 1.j*d2_o_i(_a))
                d2_A = np.abs(d2_A_r(_a) + 1.j*d2_A_i(_a))

                # Get the larger of the two a-curvatures
                d2 = np.max([d2_o, d2_A])

                # This combination has units of a. The numerator
                # is an empirical fudge factor
                _delta_a = 0.05/np.sqrt(d2)

            self.delta_a_prop.append(_delta_a)

            # Make sure it's between our min and max allowed step size
            _delta_a = np.max([self.delta_a_min, _delta_a])
            _delta_a = np.min([self.delta_a_max, _delta_a])

            _a = _a + _delta_a

            # Make sure we get the end point
            if (_a > self.a_max):
                _a = self.a_max

            omega_guess = interp_o_r(_a) + 1.j*interp_o_i(_a)
            A0          = interp_A_r(_a) + 1.j*interp_A_i(_a)

        return _a, omega_guess, A0

    def build_interps(self):
        """Build interpolating functions for omega(a) and A(a).

        This is automatically called at the end of :meth:`do_find_sequence`.
        """

        # TODO do we want to allow extrapolation?

        k=3 # cubic

        # Sadly, UnivariateSpline does not work on complex data
        self._interp_o_r = interpolate.UnivariateSpline(
            self.a, np.real(self.omega),
            s=0, # No smoothing!
            k=k, ext=0)

        self._interp_o_i = interpolate.UnivariateSpline(
            self.a, np.imag(self.omega),
            s=0, # No smoothing!
            k=k, ext=0)

        self._interp_A_r = interpolate.UnivariateSpline(
            self.a, np.real(self.A),
            s=0, # No smoothing!
            k=k, ext=0)

        self._interp_A_i = interpolate.UnivariateSpline(
            self.a, np.imag(self.A),
            s=0, # No smoothing!
            k=k, ext=0)

    def __call__(self, a, store=False, interp_only=False, resolve_if_found=False):
        """Solve for omega, A, and C[] at a given spin a.

        This uses the interpolants, based on the solved sequence, for
        initial guesses of omega(a) and A(a).

        Parameters
        ----------
        a: float
          Value of spin, 0 <= a < 1.

        store: bool, optional [default: False]
          Whether or not to save newly solved data in sequence.
          Warning, this can produce a slowdown if a lot of data
          needs to be moved.

        interp_only: bool, optional [default: False]
          If False, use the Leaver solver to polish the interpolated
          guess. If True, just use the interpolated guess.

        resolve_if_found: bool, optional [default: False]
          If False, and the value of `a` is found in the sequence, the
          previously-found solution is returned. If True, the Leaver
          solver will be used to polish the root with the current
          parameters for the solver.

        Returns
        -------
        complex, complex, complex ndarray
          The first element of the tuple is omega. The second element
          of the tuple is A. The third element of the tuple is the
          array of complex spherical-spheroidal decomposition
          coefficients. For documentation on the format of the
          spherical-spheroidal decomposition coefficient array, see
          :mod:`qnm.angular` or
          :func:`qnm.angular.C_and_sep_const_closest`.
        """

        # TODO take parameter of whether to solve at guess or not
        if not ((isinstance(a, float) or isinstance(a, int))
                and (a >= 0.) and (a < 1.)):
            raise ValueError("a={} is not a float in the range [0,1)".format(a))

        # Validate inputs
        if interp_only:
            if store:
                logging.warn("store=True will be ignored when "
                             "interp_only=True.  Setting to False.")
                store = False
            if resolve_if_found:
                logging.warn("resolve_if_found=True will be ignored when "
                             "interp_only=True.  Setting to False.")
                resolve_if_found = False

        # If this was a previously computed value, and the user
        # doesn't want to re-solve, just return the earlier results
        if ((not resolve_if_found) and (a in self.a)):
            a_ind = self.a.index(a)

            return self.omega[a_ind], self.A[a_ind], self.C[a_ind]

        # TODO Make sure that interpolants have been built

        o_r = self._interp_o_r(a)
        o_i = self._interp_o_i(a)

        A_r = self._interp_A_r(a)
        A_i = self._interp_A_i(a)

        omega_guess = complex(o_r, o_i)
        A_guess     = complex(A_r, A_i)

        if interp_only:
            # Return the interpolated (or extrapolated) values.
            # For the C coefficients, just get the eigenvector at this
            # omega_guess closest to this A_guess
            c = a * omega_guess
            A_guess, C_guess = \
                C_and_sep_const_closest(A_guess, self.s, c,
                                        self.m, self.l_max)
            return omega_guess, A_guess, C_guess

        self.solver.set_params(a=a, omega_guess=omega_guess,
                               A_closest_to=A_guess)

        result = self.solver.do_solve()

        if (result is None):
            raise optimize.nonlin.NoConvergence('Failed to find '
                                                'QNM in sequence '
                                                'at a={}'.format(a))

        cf_err, n_frac = self.solver.get_cf_err()

        # If we got here, then this new value of a was not already in
        # the list
        if store:
            if (a in self.a):
                insert_ind = self.a.index(a)
            else:
                # Where do we want to insert? Before the first point with
                # a larger spin
                try:
                    insert_ind = next(i for i, _a in
                                      enumerate( self.a ) if _a > a)
                except StopIteration:
                    insert_ind = len(self.a)

            self.a.insert(insert_ind, a)
            self.omega.insert(insert_ind, result)
            self.A.insert(insert_ind, self.solver.A)
            self.C.insert(insert_ind, self.solver.C)
            self.cf_err.insert(insert_ind, cf_err)
            self.n_frac.insert(insert_ind, n_frac)

        return result, self.solver.A, self.solver.C

    def __repr__(self):
    # "The goal of __str__ is to be readable; the goal of __repr__ is to be unambiguous." --- stackoverflow
        from textwrap import dedent

        rep = """<{} with s={}, l={}, m={}, n={},
             l_max={},
             tol={},
             Nr={}, Nr_min={}, Nr_max={},
             with values at
             a=[{}, ... <{}> ..., {}]>"""
        rep = rep.format(type(self).__name__,
                         str(self.s), str(self.l),
                         str(self.m), str(self.n),
                         str(self.l_max),
                         str(self.tol),
                         str(self.Nr), str(self.Nr_min), str(self.Nr_max),
                         str(self.a[0]), str(len(self.a)-2), str(self.a[-1]))

        return dedent(rep)



import pytest
import qnm
import numpy as np
try:
    from pathlib import Path # py 3
except ImportError:
    from pathlib2 import Path # py 2

class QnmTestDownload(object):
    """
    Base class so that each test will automatically download_data
    """
    @classmethod
    def setup_class(cls):
        """
        Download the data when setting up the test class.
        """
        qnm.download_data()

class TestQnmFileOps(QnmTestDownload):
    def test_cache_file_operations(self):
        """Test file operations and downloading the on-disk cache.
        """

        print("Downloading with overwrite=True")
        qnm.cached.download_data(overwrite=True)
        print("Clearing disk cache but not tarball")
        qnm.cached._clear_disk_cache(delete_tarball=False)
        print("Decompressing tarball")
        qnm.cached._decompress_data()

class TestQnmOneMode(QnmTestDownload):
    def test_one_mode(self):
        """
        An example of a test
        """
        grav_220 = qnm.modes_cache(s=-2,l=2,m=2,n=0)
        omega, A, C = grav_220(a=0.68)
        assert np.allclose(omega, (0.5239751042900845 - 0.08151262363119974j))

class TestQnmNewLeaverSolver(QnmTestDownload):
    def test_compare_old_new_Leaver(self):
        """ Check consistency between old and new Leaver solvers """
        from qnm.radial import leaver_cf_inv_lentz_old, leaver_cf_inv_lentz
        old = leaver_cf_inv_lentz_old(omega=.4 - 0.2j, a=0.02, s=-2, m=2, A=4.+0.j, n_inv=0)
        new = leaver_cf_inv_lentz(omega=.4 - 0.2j, a=0.02, s=-2, m=2, A=4.+0.j, n_inv=0)
        assert np.all([old[i] == new[i] for i in range(3)])

class TestQnmSolveInterface(QnmTestDownload):
    """
    Test the various interface options for solving
    """

    def test_interp_only(self):
        """Check that we get reasonable values (but not identical!)
        with just interpolation.
        """

        grav_220 = qnm.modes_cache(s=-2,l=2,m=2,n=0)
        a = 0.68
        assert a not in grav_220.a

        omega_int, A_int, C_int = grav_220(a=a, interp_only=True)
        omega_sol, A_sol, C_sol = grav_220(a=a, interp_only=False, store=False)

        assert np.allclose(omega_int, omega_sol) and not np.equal(omega_int, omega_sol)
        assert np.allclose(A_int, A_sol) and not np.equal(A_int, A_sol)
        assert np.allclose(C_int, C_sol) and not all(np.equal(C_int, C_sol))

    def test_store_a(self):
        """Check that the option store=True updates a spin sequence"""

        grav_220 = qnm.modes_cache(s=-2,l=2,m=2,n=0)

        old_n = len(grav_220.a)
        k = int(old_n/2)

        new_a = 0.5 * (grav_220.a[k] + grav_220.a[k+1])

        assert new_a not in grav_220.a

        _, _, _ = grav_220(new_a, store=False)
        n_1 = len(grav_220.a)
        assert old_n == n_1

        _, _, _ = grav_220(new_a, store=True)
        n_2 = len(grav_220.a)
        assert n_2 == n_1 + 1

    def test_resolve(self):
        """Test that option resolve_if_found=True really does a new
        solve"""

        grav_220 = qnm.modes_cache(s=-2,l=2,m=2,n=0)

        n = len(grav_220.a)
        k = int(n/2)
        a = grav_220.a[k]

        omega_old, A_old, C_old = grav_220(a=a, resolve_if_found=False)

        old_tol = grav_220.solver.tol
        old_cf_tol = grav_220.solver.cf_tol

        grav_220.solver.set_params(tol = old_tol * 0.1, cf_tol = old_cf_tol * 0.1)

        omega_new, A_new, C_new = grav_220(a=a, resolve_if_found=True)

        assert np.allclose(omega_new, omega_old) and not np.equal(omega_new, omega_old)
        assert np.allclose(A_new, A_old) and not np.equal(A_new, A_old)
        assert np.allclose(C_new, C_old) and not all(np.equal(C_new, C_old))

class TestQnmBuildCache(QnmTestDownload):
    def test_build_cache(self):
        """Check the default cache-building functionality"""

        qnm.cached._clear_disk_cache(delete_tarball=False)
        qnm.modes_cache.seq_dict = {}
        qnm.cached.build_package_default_cache(qnm.modes_cache)
        assert 860 == len(qnm.modes_cache.seq_dict.keys())
        qnm.modes_cache.write_all()
        cache_data_dir = qnm.cached.get_cachedir() / 'data'

        # Magic number, default num modes is 860
        assert 860 == len(list(cache_data_dir.glob('*.pickle')))



""" Find a nearby root of the coupled radial/angular Teukolsky equations.

TODO Documentation.
"""

from __future__ import division, print_function, absolute_import

import logging

import numpy as np
from scipy import optimize

from .angular import sep_const_closest, C_and_sep_const_closest
from . import radial

# TODO some documentation here, better documentation throughout

class NearbyRootFinder(object):
    """Object to find and store results from simultaneous roots of
    radial and angular QNM equations, following the
    Leaver and Cook-Zalutskiy approach.

    Parameters
    ----------
    a: float [default: 0.]
      Dimensionless spin of black hole, 0 <= a < 1.

    s: int [default: -2]
      Spin of field of interest

    m: int [default: 2]
      Azimuthal number of mode of interest

    A_closest_to: complex [default: 4.+0.j]
      Complex value close to desired separation constant. This is
      intended for tracking the l-number of a sequence starting
      from the analytically-known value at a=0

    l_max: int [default: 20]
      Maximum value of l to include in the spherical-spheroidal
      matrix for finding separation constant and mixing
      coefficients. Must be sufficiently larger than l of interest
      that angular spectral method can converge. The number of
      l's needed for convergence depends on a.

    omega_guess: complex [default: .5-.5j]
      Initial guess of omega for root-finding

    tol: float [default: sqrt(double epsilon)]
      Tolerance for root-finding omega

    cf_tol: float [defailt: 1e-10]
      Tolerance for continued fraction calculation

    n_inv: int [default: 0]
      Inversion number of radial infinite continued fraction,
      which selects overtone number of interest

    Nr: int [default: 300]
      Truncation number of radial infinite continued
      fraction. Must be sufficiently large for convergence.

    Nr_min: int [default: 300]
      Floor for Nr (for dynamic control of Nr)

    Nr_max: int [default: 4000]
      Ceiling for Nr (for dynamic control of Nr)

    r_N: complex [default: 1.]
      Seed value taken for truncation of infinite continued
      fraction. UNUSED, REMOVE

    """

    def __init__(self, *args, **kwargs):

        # Set defaults before using values in kwargs
        self.a           = 0.
        self.s           = -2
        self.m           = 2
        self.A0          = 4.+0.j
        self.l_max       = 20
        self.omega_guess = .5-.5j
        self.tol         = np.sqrt(np.finfo(float).eps)
        self.cf_tol      = 1e-10
        self.n_inv       = 0
        self.Nr          = 300
        self.Nr_min      = 300
        self.Nr_max      = 4000
        self.r_N         = 1.

        self.set_params(**kwargs)

    def set_params(self, *args, **kwargs):
        """Set the parameters for root finding. Parameters are
        described in the class documentation. Finally calls
        :meth:`clear_results`.
        """

        # TODO This violates DRY, do better.
        self.a           = kwargs.get('a',            self.a)
        self.s           = kwargs.get('s',            self.s)
        self.m           = kwargs.get('m',            self.m)
        self.A0          = kwargs.get('A_closest_to', self.A0)
        self.l_max       = kwargs.get('l_max',        self.l_max)
        self.omega_guess = kwargs.get('omega_guess',  self.omega_guess)
        self.tol         = kwargs.get('tol',          self.tol)
        self.cf_tol      = kwargs.get('cf_tol',       self.cf_tol)
        self.n_inv       = kwargs.get('n_inv',        self.n_inv)
        self.Nr          = kwargs.get('Nr',           self.Nr)
        self.Nr_min      = kwargs.get('Nr_min',       self.Nr_min)
        self.Nr_max      = kwargs.get('Nr_max',       self.Nr_max)
        self.r_N         = kwargs.get('r_N',          self.r_N)

        # Optional pole factors
        self.poles       = np.array([])

        # TODO: Check that values make sense

        self.clear_results()

    def clear_results(self):
        """Clears the stored results from last call of :meth:`do_solve`"""

        self.solved  = False
        self.opt_res = None

        self.omega = None
        self.A     = None
        self.C     = None

        self.cf_err = None
        self.n_frac = None

        self.poles = np.array([])


    def __call__(self, x):
        """Internal function for usage with optimize.root, for an
        instance of this class to act like a function for
        root-finding. optimize.root only works with reals so we pack
        and unpack complexes into float[2]
        """

        omega = x[0] + 1.j*x[1]
        # oblateness parameter
        c     = self.a * omega
        # Separation constant at this a*omega
        A     = sep_const_closest(self.A0, self.s, c, self.m,
                                  self.l_max)

        # We are trying to find a root of this function:
        # inv_err = radial.leaver_cf_trunc_inversion(omega, self.a,
        #                                            self.s, self.m, A,
        #                                            self.n_inv,
        #                                            self.Nr, self.r_N)

        # TODO!
        # Determine the value to use for cf_tol based on
        # the Jacobian, cf_tol = |d cf(\omega)/d\omega| tol.
        inv_err, self.cf_err, self.n_frac = radial.leaver_cf_inv_lentz(omega, self.a,
                                                          self.s, self.m, A,
                                                          self.n_inv, self.cf_tol,
                                                          self.Nr_min, self.Nr_max)
        # logging.info("Lentz terminated with cf_err={}, n_frac={}".format(self.cf_err, self.n_frac))

        # Insert optional poles
        pole_factors   = np.prod(omega - self.poles)
        supp_err = inv_err / pole_factors

        return [np.real(supp_err), np.imag(supp_err)]

    def do_solve(self):
        """Try to find a root of the continued fraction equation,
        using the parameters that have been set in :meth:`set_params`."""

        # For the default (hybr) method, tol sets 'xtol', the
        # tolerance on omega.
        self.opt_res = optimize.root(self,
                                     [np.real(self.omega_guess), np.imag(self.omega_guess)],
                                     method = 'hybr', tol = self.tol)

        if (not self.opt_res.success):
            tmp_opt_res = self.opt_res
            self.clear_results()
            self.opt_res = tmp_opt_res
            return None

        self.solved = True

        self.omega = self.opt_res.x[0] + 1.j*self.opt_res.x[1]
        c = self.a * self.omega
        # As far as I can tell, scipy.linalg.eig already normalizes
        # the eigenvector to unit norm, and the coefficient with the
        # largest norm is real
        self.A, self.C = C_and_sep_const_closest(self.A0,
                                                 self.s, c,
                                                 self.m, self.l_max)

        return self.omega

    def get_cf_err(self):
        """Return the continued fraction error and the number of
        iterations in the last evaluation of the continued fraction.

        Returns
        -------
        cf_err: float

        n_frac: int
        """

        return self.cf_err, self.n_frac

    def set_poles(self, poles=[]):
        """ Set poles to multiply error function.

        Parameters
        ----------
        poles: array_like as complex numbers [default: []]

        """

        self.poles = np.array(poles).astype(complex)



""" Finding QNMs of Schwarzschild (numerically and approximately).

Schwarzschild QNMs are used as starting points for following a Kerr
QNM along a "spin sequence" in :class:`qnm.spinsequence.KerrSpinSeq`.
Most end users will not need to directly use the functions in this
module.
"""

from __future__ import print_function, division, absolute_import

from . import approx
from . import tabulated
from . import overtonesequence



# -*- coding: utf-8 -*-
"""Solve the angular Teukolsky equation via spectral decomposition.

For a given complex QNM frequency ω, the separation constant and
spherical-spheroidal decomposition are found as an eigenvalue and
eigenvector of an (infinite) matrix problem.  The interface to solving
this problem is :meth:`C_and_sep_const_closest`, which returns a
certain eigenvalue A and eigenvector C.  The eigenvector contains the
C coefficients in the equation

.. math:: {}_s Y_{\ell m}(\\theta, \phi; a\omega) = {\sum_{\ell'=\ell_{\min} (s,m)}^{\ell_\max}} C_{\ell' \ell m}(a\omega)\ {}_s Y_{\ell' m}(\\theta, \phi) \,.

Here ℓmin=max(\|m\|,\|s\|) (see :meth:`l_min`), and ℓmax can be chosen at
run time. The C coefficients are returned as a complex ndarray, with
the zeroth element corresponding to ℓmin.  You can get the associated
ℓ values by calling :meth:`ells`.
"""

from __future__ import division, print_function, absolute_import

from numba import njit
import numpy as np

# TODO some documentation here, better documentation throughout

@njit(cache=True)
def _calF(s, l, m):
    """ Eq. (52b) """

    if ((0==s) and (0 == l+1)):
        # This can only happen when solving for the mode labeled by s=0, l=0, m=0
        return 0.

    return (np.sqrt( ((l+1)**2 - m*m) / (2*l+3) / (2*l+1) )
            * np.sqrt( ( (l+1)**2  - s*s)  / (l+1)**2 ))

@njit(cache=True)
def _calG(s, l, m):
    """ Eq. (52c) """
    if (0 == l):
        return 0.

    return np.sqrt( ( l*l - m*m ) / (4*l*l - 1)) * np.sqrt(1 - s*s/l/l)

@njit(cache=True)
def _calH(s, l, m):
    """ Eq. (52d) """
    if (0 == l) or (0 == s):
        return 0.

    return - m*s/l/(l+1)

@njit(cache=True)
def _calA(s, l, m):
    """ Eq. (53a) """
    return _calF(s,l,m) * _calF(s,l+1,m)

@njit(cache=True)
def _calD(s, l, m):
    """ Eq. (53b) """
    return _calF(s,l,m) * (_calH(s,l+1,m)  + _calH(s,l,m))

@njit(cache=True)
def _calB(s, l, m):
    """ Eq. (53c) """
    return (_calF(s,l,m) * _calG(s,l+1,m)
            + _calG(s,l,m) * _calF(s,l-1,m)
            + _calH(s,l,m)**2)

@njit(cache=True)
def _calE(s, l, m):
    """ Eq. (53d) """
    return _calG(s,l,m) * (_calH(s,l-1,m) + _calH(s,l,m))

@njit(cache=True)
def _calC(s, l, m):
    """ Eq. (53e) """
    return _calG(s,l,m) * _calG(s,l-1,m)

@njit(cache=True)
def swsphericalh_A(s, l, m):
    """ Angular separation constant at a=0.

    Eq. (50). Has no dependence on m. The formula is
      A_0 = l(l+1) - s(s+1)

    Parameters
    ----------
    s: int
      Spin-weight of interest

    l: int
      Angular quantum number of interest

    m: int
      Magnetic quantum number, ignored

    Returns
    -------
    int
      Value of A(a=0) = l(l+1) - s(s+1)
    """

    return l*(l+1) - s*(s+1)

@njit(cache=True)
def M_matrix_elem(s, c, m, l, lprime):
    """ The (l, lprime) matrix element from the spherical-spheroidal
    decomposition matrix from Eq. (55).

    Parameters
    ----------
    s: int
      Spin-weight of interest

    c: complex
      Oblateness of the spheroidal harmonic

    m: int
      Magnetic quantum number

    l: int
      Angular quantum number of interest

    lprime: int
      Primed quantum number of interest

    Returns
    -------
    complex
      Matrix element M_{l, lprime}
    """

    if (lprime == l-2):
        return -c*c*_calA(s,lprime,m)
    if (lprime == l-1):
        return (-c*c*_calD(s,lprime,m)
                + 2*c*s*_calF(s,lprime,m))
    if (lprime == l  ):
        return (swsphericalh_A(s,lprime,m)
                - c*c*_calB(s,lprime,m)
                + 2*c*s*_calH(s,lprime,m))
    if (lprime == l+1):
        return (-c*c*_calE(s,lprime,m)
                + 2*c*s*_calG(s,lprime,m))
    if (lprime == l+2):
        return -c*c*_calC(s,lprime,m)

    return 0.

def give_M_matrix_elem_ufunc(s, c, m):
    """Legacy function. Gives ufunc that implements matrix elements of
    the spherical-spheroidal decomposition matrix. This function is
    used by :meth:`M_matrix_old`.

    Parameters
    ----------
    s: int
      Spin-weight of interest

    c: complex
      Oblateness of the spheroidal harmonic

    m: int
      Magnetic quantum number

    Returns
    -------
    ufunc
      Implements elements of M matrix
    """

    def elem(l, lprime):
        return M_matrix_elem(s, c, m, l, lprime)

    return np.frompyfunc(elem, 2, 1)

@njit(cache=True)
def l_min(s, m):
    """ Minimum allowed value of l for a given s, m.

    The formula is l_min = max(\|m\|,\|s\|).

    Parameters
    ----------
    s: int
      Spin-weight of interest

    m: int
      Magnetic quantum number

    Returns
    -------
    int
      l_min
    """

    return max(abs(s), abs(m))

@njit(cache=True)
def ells(s, m, l_max):
    """Vector of ℓ values in C vector and M matrix.

    The format of the C vector and M matrix is that the 0th element
    corresponds to l_min(s,m) (see :meth:`l_min`).

    Parameters
    ----------
    s: int
      Spin-weight of interest

    m: int
      Magnetic quantum number

    l_max: int
      Maximum angular quantum number

    Returns
    -------
    int ndarray
      Vector of ℓ values, starting from l_min
    """

    return np.arange(l_min(s,m), l_max+1)

def M_matrix_old(s, c, m, l_max):
    """Legacy function. Same as :meth:`M_matrix` except trying to be cute
    with ufunc's, requiring scope capture with temp func inside
    :meth:`give_M_matrix_elem_ufunc`, which meant that numba could not
    speed up this method. Remains here for testing purposes. See
    documentation for :meth:`M_matrix` parameters and return value.
    """

    _ells = ells(s, m, l_max)

    uf = give_M_matrix_elem_ufunc(s, c, m)

    return uf.outer(_ells,_ells).astype(complex)

@njit(cache=True)
def M_matrix(s, c, m, l_max):
    """Spherical-spheroidal decomposition matrix truncated at l_max.

    Parameters
    ----------
    s: int
      Spin-weight of interest

    c: complex
      Oblateness of the spheroidal harmonic

    m: int
      Magnetic quantum number

    l_max: int
      Maximum angular quantum number

    Returns
    -------
    complex ndarray
      Decomposition matrix
    """

    _ells = ells(s, m, l_max)

    M = np.empty((len(_ells),len(_ells)), dtype=np.complex128)

    for i in range(len(_ells)):
        for j in range(len(_ells)):
            M[i,j] = M_matrix_elem(s, c, m, _ells[i], _ells[j])

    return M

def sep_consts(s, c, m, l_max):
    """Finds eigenvalues of decomposition matrix, i.e. the separation
    constants, As.

    Parameters
    ----------
    s: int
      Spin-weight of interest

    c: complex
      Oblateness of spheroidal harmonic

    m: int
      Magnetic quantum number

    l_max: int
      Maximum angular quantum number

    Returns
    -------
    complex ndarray
      Eigenvalues of spherical-spheroidal decomposition matrix
    """

    return np.linalg.eigvals(M_matrix(s, c, m, l_max))

def sep_const_closest(A0, s, c, m, l_max):
    """Gives the separation constant that is closest to A0.

    Parameters
    ----------
    A0: complex
      Value close to the desired separation constant.

    s: int
      Spin-weight of interest

    c: complex
      Oblateness of spheroidal harmonic

    m: int
      Magnetic quantum number

    l_max: int
      Maximum angular quantum number

    Returns
    -------
    complex
      Separation constant that is the closest to A0.
    """

    As = sep_consts(s, c, m, l_max)
    i_closest = np.argmin(np.abs(As-A0))
    return As[i_closest]

def C_and_sep_const_closest(A0, s, c, m, l_max):
    """Get a single eigenvalue and eigenvector of decomposition
    matrix, where the eigenvalue is closest to some guess A0.

    Parameters
    ----------
    A0: complex
      Value close to the desired separation constant.

    s: int
      Spin-weight of interest

    c: complex
      Oblateness of spheroidal harmonic

    m: int
      Magnetic quantum number

    l_max: int
      Maximum angular quantum number

    Returns
    -------
    complex, complex ndarray
      The first element of the tuple is the eigenvalue that is closest
      in value to A0. The second element of the tuple is the
      corresponding eigenvector.  The 0th element of this ndarray
      corresponds to :meth:`l_min`.
    """

    As, Cs = np.linalg.eig(M_matrix(s, c, m, l_max))
    i_closest = np.argmin(np.abs(As-A0))
    return As[i_closest], Cs[:,i_closest]
